# coding: utf-8
import logging
import time
import typing
from typing import Union, Dict, List, Tuple, Optional, Any

import numpy as np
from PIL import Image
from selenium.types import WaitExcTypes

from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import WebDriverException, TimeoutException, NoSuchElementException
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.common.by import By
from appium.webdriver import WebElement
from appium.webdriver.common.mobileby import MobileBy
from appium.webdriver.common.appiumby import AppiumBy

from appium_extended.appium_base import AppiumBase


class AppiumGet(AppiumBase):
    """
    Класс расширяющий Appium.
    Обеспечивает получение чего-либо со страницы.
    """

    def __init__(self, logger: logging.Logger):
        super().__init__(logger=logger)

    def _get_element(self,
                     locator: Union[Tuple, WebElement, 'WebElementExtended', Dict[str, str], str] = None,
                     by: Union[MobileBy, AppiumBy, By, str] = None,
                     value: Union[str, Dict, None] = None,
                     timeout_elem: float = 10,
                     timeout_method: float = 600,
                     elements_range: Union[Tuple, List[WebElement], Dict[str, str], None] = None,
                     contains: bool = True,
                     poll_frequency: float = 0.5,
                     ignored_exceptions: typing.Optional[WaitExcTypes] = None
                     ) -> \
            Union[WebElement, None]:
        """
        Метод обеспечивает поиск элемента в текущей DOM структуре.
        Должен принимать либо локатор, либо значения by и value.

        Args:
            locator (Union[Tuple, WebElement, 'WebElementExtended', Dict[str, str], str], optional):
                Определяет локатор элемента.
                Tuple - локатор в виде кортежа из двух строковых элементов,
                    где первый это стратегия поиска, а второй это селектор,
                    например ("id", "android.widget.ProgressBar").
                Dict - локатор в виде словаря атрибутов и их значений искомого элемента,
                    например {'text': 'foo', 'displayed' : 'true', 'enabled': 'true'}.

            by (Union[MobileBy, AppiumBy, By, str], optional):
                Тип локатора для поиска элемента (всегда в связке с value).
                Как в стандартном методе driver.find_element.
            value (Union[str, Dict, None], optional):
                Значение локатора или словарь аргументов, если используется AppiumBy.XPATH.
            timeout_elem (int, optional):
                Время ожидания элемента. По умолчанию 10 секунд.
            timeout_method (int, optional):
                Время ожидания метода поиска элемента. По умолчанию 600 секунд.
            elements_range (Union[Tuple, List[WebElement], Dict[str, str], None], optional):
                Ограничивает поиск элемента в указанном диапазоне (для поиска по изображению).
            contains (bool, optional):
                Для поиска по dict и атрибуту 'text',
                ищет элемент содержащий фрагмент значения если True
                и по строгому соответствию если False.
                По умолчанию True.

        Usages:
            element = app._get_element(locator=("id", "foo"))
            element = app._get_element(element)
            element = app._get_element(locator={'text': 'foo'}, contains=True)
            element = app._get_element(locator='/path/to/file/image.png')
            element = app._get_element(by="id", value="backButton")
            element = app._get_element(by=MobileBy.ID, value="backButton")

        Returns:
            Union[WebElement, None]: Возвращает WebElement, если элемент найден, иначе None.
        """
        # Проверка и подготовка аргументов
        if (not locator) and (not by or not value):
            self.logger.error(f"Некорректные аргументы!\n"
                              f"{locator=}\n"
                              f"{by=}\n"
                              f"{value=}\n"
                              f"{timeout_elem=}\n")
            return None
        if not locator and (by and value):
            locator = (by, value)
        if locator is None:
            return None

        # Объявление стратегии поиска элементов
        locator_handler = {
            # возвращает себя же
            WebElement: self.helpers.handle_webelement_locator,
            # возвращает себя же
            'WebElementExtended': self.helpers.handle_webelement_locator,
            # составляет локатор типа tuple из словаря с атрибутами искомого элемента
            dict: self.helpers.handle_dict_locator,
            # производит поиск элементов по фрагменту изображения, возвращает список элементов
            str: self.helpers.handle_string_locator,
        }

        # Цикл подготовки локатора и поиска элементов
        start_time = time.time()
        while not isinstance(locator, WebElement) and time.time() - start_time < timeout_method:
            # Выявление типа данных локатора для его подготовки
            locator_type = type(locator)
            # Если локатор типа tuple, то выполняется извлечение элементов
            if isinstance(locator, tuple):
                wait = WebDriverWait(driver=self.driver, timeout=timeout_elem,
                                     poll_frequency=poll_frequency, ignored_exceptions=ignored_exceptions)
                try:
                    element = wait.until(EC.presence_of_element_located(locator))
                    return element
                except NoSuchElementException:
                    return None
                except TimeoutException as error:
                    self.logger.debug(f"Элемент не обнаружен!\n"
                                      f"{locator=}\n"
                                      f"{timeout_elem=}\n\n" +
                                      "{}\n".format(error))
                    self.logger.debug("page source ", self.driver.page_source)
                    return None
                except WebDriverException as error:
                    self.logger.debug(f"Элемент не обнаружен!\n"
                                      f"{locator=}\n"
                                      f"{timeout_elem=}\n\n" +
                                      "{}\n".format(error))
                    self.logger.debug("page source ", self.driver.page_source)
                    return None
            # Выполнение подготовки локатора
            handler = locator_handler.get(locator_type)
            if locator is None:
                return None
            locator = handler(locator=locator, timeout=timeout_elem, elements_range=elements_range, contains=contains)
        # Подбирает результат после поиска по изображению
        if isinstance(locator, WebElement):
            return locator
        self.logger.error(f"Что-то пошло не так\n"
                          f"{locator=}\n"
                          f"{by=}\n"
                          f"{value=}\n"
                          f"{timeout_elem=}\n"
                          f"{timeout_method=}\n")
        return None

    def _get_elements(self,
                      locator: Union[Tuple, List[WebElement], Dict[str, str], str] = None,
                      by: Union[MobileBy, AppiumBy, By, str] = None,
                      value: Union[str, Dict, None] = None,
                      timeout_elements: int = 10,
                      timeout_method: int = 600,
                      elements_range: Union[Tuple, List[WebElement], Dict[str, str], None] = None,
                      contains: bool = True,
                      poll_frequency: float = 0.5,
                      ignored_exceptions: typing.Optional[WaitExcTypes] = None
                      ) -> \
            Union[List[WebElement], None]:
        """
        Метод обеспечивает поиск элементов в текущей DOM структуре.
        Должен принять либо локатор, либо by и value.

        Args:
            locator (Union[Tuple, List[WebElement], Dict[str, str], str], optional):
                Определяет локатор элементов.
                Tuple - локатор в виде кортежа из двух строковых элементов,
                    где первый это стратегия поиска, а второй это селектор,
                    например ("id", "android.widget.ProgressBar").
                Dict - локатор в виде словаря атрибутов и их значений искомого элемента,
                    например {'text': 'foo', 'displayed' : 'true', 'enabled': 'true'}.
            by (Union[MobileBy, AppiumBy, By, str], optional):
                Тип локатора для поиска элементов (всегда в связке с value).
                Как в стандартном методе driver.find_element.
            value (Union[str, Dict, None], optional):
                Значение локатора или словарь аргументов, если используется XPATH.
            timeout_elements (int, optional):
                Время ожидания элементов. По умолчанию 10 секунд.
            timeout_method (int, optional):
                Время ожидания метода поиска элементов. По умолчанию 600 секунд.
            elements_range (Union[Tuple, List[WebElement], Dict[str, str], None], optional):
                Ограничивает поиск элементов в указанном диапазоне.
            contains (bool, optional):
                Для поиска по dict и атрибуту 'text',
                True - ищет элемент содержащий фрагмент значения,
                False - по строгому соответствию.
                По умолчанию True.

        Usages:
            elements = app._get_elements(locator=("id", "foo"))
            elements = app._get_elements(locator={'text': 'foo'})
            elements = app._get_elements(locator='/path/to/file/pay_agent.png')
            elements = app._get_elements(by="id", value="ru.sigma.app.debug:id/backButton")
            elements = app._get_elements(by=MobileBy.ID, value="ru.sigma.app.debug:id/backButton")

        Raises:
            WebDriverException: Если произошла ошибка при взаимодействии с WebDriver.

        Returns:
            Union[List[WebElement], None]: Возвращает список WebElement'ов, если элементы найдены, иначе None.
        """
        # Проверка и подготовка аргументов
        if not locator and (not by or not value):
            self.logger.error(f"Некорректные аргументы!\n"
                              f"{locator=}\n"
                              f"{by=}\n"
                              f"{value=}\n"
                              f"{timeout_elements=}\n"
                              f"{timeout_method=}\n\n" +
                              f"{poll_frequency=}\n" +
                              f"{ignored_exceptions=}\n")
            return None
        if not locator and (by and value):
            locator = (by, value)
        if locator is None:
            return None

        # Объявление стратегии поиска элементов
        locator_handler = {
            # подразумевается список элементов, возвращает себя же
            list: self.helpers.handle_webelement_locator_elements,
            # составляет локатор типа tuple из словаря с атрибутами искомого элемента
            dict: self.helpers.handle_dict_locator_elements,
            # производит поиск элементов по фрагменту изображения, возвращает список элементов
            str: self.helpers.handle_string_locator_elements,
        }

        # Цикл подготовки локатора и поиска элементов
        start_time = time.time()
        while not isinstance(locator, list) and time.time() - start_time < timeout_method:
            # Выявление типа данных локатора для его подготовки
            locator_type = type(locator)
            # Если локатор типа tuple, то выполняется извлечение элементов
            if isinstance(locator, tuple):
                wait = WebDriverWait(driver=self.driver, timeout=timeout_elements,
                                     poll_frequency=poll_frequency, ignored_exceptions=ignored_exceptions)
                try:
                    element = wait.until(EC.presence_of_all_elements_located(locator))
                    return element
                except WebDriverException as error:
                    self.logger.debug(f"Элемент не обнаружен!\n"
                                      f"{locator=}\n"
                                      f"{by=}\n"
                                      f"{value=}\n"
                                      f"{timeout_elements=}\n"
                                      f"{timeout_method=}\n\n" +
                                      f"{poll_frequency=}\n" +
                                      f"{ignored_exceptions=}\n" +
                                      "{}\n".format(error))
                    return None
            # Выполнение подготовки локатора
            handler = locator_handler.get(locator_type)
            locator = handler(locator=locator,
                              timeout=timeout_elements,
                              elements_range=elements_range,
                              contains=contains)
        # Подбирает результат после поиска по изображению
        if isinstance(locator, list):
            return locator
        self.logger.debug(f"\nЧто-то пошло не так\n"
                          f"{locator=}\n"
                          f"{by=}\n"
                          f"{value=}\n"
                          f"{timeout_elements=}\n"
                          f"{timeout_method=}\n")
        return None

    def _get_image_coordinates(self,
                               image: Union[bytes, np.ndarray, Image.Image, str],
                               full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                               threshold: float = 0.7,
                               ) -> Union[Tuple[int, int, int, int], None]:
        return self.helpers.get_image_coordinates(image=image, full_image=full_image, threshold=threshold)

    def _get_inner_image_coordinates(self,
                                     outer_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                     inner_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                     threshold: float = 0.9) -> \
            Union[Tuple[int, int, int, int], None]:
        return self.helpers.get_inner_image_coordinates(outer_image_path=outer_image_path,
                                                       inner_image_path=inner_image_path,
                                                       threshold=threshold)

    def _get_text_coordinates(self,
                              text: str,
                              language: str = 'rus',
                              image: Union[bytes, str, Image.Image, np.ndarray] = None, ) -> Optional[tuple[int, ...]]:
        return self.helpers.get_text_coordinates(text=text, language=language, image=image)

    def _get_screenshot_as_base64_decoded(self, scrcpy: bool = False):
        return self.helpers._get_screenshot_as_base64_decoded(scrcpy=scrcpy)



# coding: utf-8
import logging
from typing import Optional

from selenium.common.exceptions import WebDriverException

from appium_extended.appium_get import AppiumGet


class AppiumTap(AppiumGet):
    """
    Класс работы с Appium.
    Обеспечивает tap
    """

    def __init__(self, logger: logging.Logger):
        super().__init__(logger=logger)

    def _tap(self,
             x: int,
             y: int,
             duration: Optional[int] = None,
             ) -> bool:
        """
        Выполняет касание на экране устройства в указанных координатах (x, y) или на основе изображения.

        Аргументы:
        - x (int): Горизонтальная координата X точки касания на экране устройства.
        - y (int): Вертикальная координата Y точки касания на экране устройства.
        - duration (Optional[int]): Длительность касания в миллисекундах (по умолчанию None).

        Возвращает:
        - True, если касание выполнено успешно.
        - False, если произошла ошибка при выполнении касания.
        """
        position = (x, y)  # Координаты для касания

        try:
            self.driver.tap(positions=[position], duration=duration)  # Выполняем касание на указанных координатах
        except WebDriverException as e:
            self.logger.error("_tap() WebDriverException "
                              f"{x=}, {y=} "
                              f"{duration=}")
            self.logger.error(e)
            return False
        return True



# coding: utf-8
import time
import typing
from typing import Union, Tuple, Dict, Optional

from appium.webdriver import WebElement
from selenium.common.exceptions import StaleElementReferenceException, NoSuchElementException, TimeoutException, \
    WebDriverException
from selenium.types import WaitExcTypes

from appium_extended_web_element.web_element_is import WebElementIs


class WebElementScroll(WebElementIs):
    """
    Класс для выполнения действий прокрутки элемента.
    Наследуется от класса WebElementGet.
    """

    def __init__(self, base, element_id):
        super().__init__(base=base, element_id=element_id)

    def _scroll_down(self,
                     locator: Union[Tuple, 'WebElementExtended', Dict[str, str], str] = None,
                     duration: int = None,
                     timeout_elements: float = 10.0,
                     timeout_method: float = 600.0,
                     elements_range: Union[tuple, list[WebElement], dict[str, str], None] = None,
                     contains: bool = True,
                     poll_frequency: float = 0.5,
                     ignored_exceptions: Optional[typing.Iterable[typing.Type[Exception]]] = None
                     ) -> bool:
        """
        Прокручивает элемент вниз от нижнего дочернего элемента до верхнего дочернего элемента родительского элемента.

        Args:
            locator (Union[Tuple, WebElement, Dict[str, str], str], optional): Локатор или элемент для прокрутки (за что крутить).
            duration (int, optional): Продолжительность прокрутки в миллисекундах (по умолчанию: None).

        Returns:
            bool: True, если прокрутка выполнена успешно.

        """
        try:
            recycler = self

            # # Проверка, является ли элемент прокручиваемым
            # if recycler.get_attribute('scrollable') != 'true':
            #     self.logger.error("Элемент не крутиться")
            #     return False

            # Если локатор для прокрутки не указан, используется локатор первого дочернего элемента
            if not locator:
                locator = {'class': self._get_first_child_class()}

            # Получение верхнего и нижнего дочерних элементов родительского элемента
            top_center_child = self._get_top_center_child_from_parent(locator=locator,
                                                                      timeout_elements=timeout_elements,
                                                                      timeout_method=timeout_method,
                                                                      elements_range=elements_range,
                                                                      contains=contains,
                                                                      poll_frequency=poll_frequency,
                                                                      ignored_exceptions=ignored_exceptions,
                                                                      )
            bottom_center_child = self._get_bottom_center_child_from_parent(locator=locator,
                                                                            timeout_elements=timeout_elements,
                                                                            timeout_method=timeout_method,
                                                                            elements_range=elements_range,
                                                                            contains=contains,
                                                                            poll_frequency=poll_frequency,
                                                                            ignored_exceptions=ignored_exceptions,
                                                                            )

            # Прокрутка вниз от нижнего дочернего элемента до верхнего дочернего элемента родительского элемента
            self.driver.scroll(origin_el=bottom_center_child, destination_el=top_center_child, duration=duration)
            return True
        except NoSuchElementException:
            return False
        except StaleElementReferenceException:
            return False
        except TimeoutException:
            return False

    def _scroll_up(self,
                   locator: Union[Tuple, 'WebElementExtended', Dict[str, str], str] = None,
                   duration: int = None,
                   timeout_elements: float = 10.0,
                   timeout_method: float = 600.0,
                   elements_range: Union[tuple, list[WebElement], dict[str, str], None] = None,
                   contains: bool = True,
                   poll_frequency: float = 0.5,
                   ignored_exceptions: Optional[typing.Iterable[typing.Type[Exception]]] = None
                   ) -> bool:
        """
        Прокручивает элемент вверх от верхнего дочернего элемента до нижнего дочернего элемента родительского элемента.

        Args:
            locator (Union[Tuple, WebElement, Dict[str, str], str], optional): Локатор или элемент для прокрутки (за что крутить).
            duration (int, optional): Продолжительность прокрутки в миллисекундах (по умолчанию: None).

        Returns:
            bool: True, если прокрутка выполнена успешно.

        """
        try:
            recycler = self

            # # Проверка, является ли элемент прокручиваемым
            # if recycler.get_attribute('scrollable') != 'true':
            #     self.logger.error("Элемент не крутиться")
            #     return False

            # Если локатор для прокрутки не указан, используется локатор первого дочернего элемента
            if not locator:
                locator = {'class': self._get_first_child_class()}

            # Получение верхнего и нижнего дочерних элементов родительского элемента
            top_center_child = self._get_top_center_child_from_parent(locator=locator,
                                                                      timeout_elements=timeout_elements,
                                                                      timeout_method=timeout_method,
                                                                      elements_range=elements_range,
                                                                      contains=contains,
                                                                      poll_frequency=poll_frequency,
                                                                      ignored_exceptions=ignored_exceptions,
                                                                      )
            bottom_center_child = self._get_bottom_center_child_from_parent(locator=locator,
                                                                            timeout_elements=timeout_elements,
                                                                            timeout_method=timeout_method,
                                                                            elements_range=elements_range,
                                                                            contains=contains,
                                                                            poll_frequency=poll_frequency,
                                                                            ignored_exceptions=ignored_exceptions,
                                                                            )

            # Прокрутка вверх от верхнего дочернего элемента до нижнего дочернего элемента родительского элемента
            self.driver.scroll(origin_el=top_center_child, destination_el=bottom_center_child, duration=duration)
            return True
        except NoSuchElementException:
            return False
        except StaleElementReferenceException:
            return False
        except TimeoutException:
            return False

    def _scroll_to_bottom(self,
                          locator: Union[Tuple, WebElement, Dict[str, str], str] = None,
                          timeout_method: int = 120) -> bool:
        """
        Прокручивает элемент вниз до упора.

        Args:
            locator (Union[Tuple, WebElement, Dict[str, str], str], optional): Локатор или элемент для прокрутки (за что крутить).
            timeout_method (int, optional): Время ожидания элемента в секундах (по умолчанию: 120).

        Returns:
            bool: True, если прокрутка выполнена успешно.

        """
        recycler = self

        # # Проверка, является ли элемент прокручиваемым
        # if recycler.get_attribute('scrollable') != 'true':
        #     self.logger.error("Элемент не крутиться")
        #     return False

        # Если локатор для прокрутки не указан, используется локатор первого дочернего элемента
        if not locator:
            locator = {'class': self._get_first_child_class()}

        last_child = None
        start_time = time.time()

        # Прокрутка вниз до упора
        while time.time() - start_time < timeout_method:
            try:
                child = self._get_element(locator=locator)
                if child == last_child:
                    return True
                last_child = child
                self._scroll_down(locator=locator)
            except StaleElementReferenceException:
                continue
        self.logger.error("_scroll_to_bottom(): Неизвестная ошибка")
        return False

    def _scroll_to_top(self,
                       locator: Union[Tuple, WebElement, Dict[str, str], str],
                       duration: int = 0.5,
                       timeout_method: int = 120,
                       timeout_elem: float = 10.0,
                       timeout_elements: float = 10.0,
                       elements_range: Union[tuple, list[WebElement], dict[str, str], None] = None,
                       contains: bool = True,
                       poll_frequency: float = 0.5,
                       ignored_exceptions: Optional[typing.Iterable[typing.Type[Exception]]] = None
                       ) -> bool:
        """
        Прокручивает элемент вверх до упора.

        Args:
            locator (Union[Tuple, WebElement, Dict[str, str], str]): Локатор или элемент для прокрутки (за что крутить).
            timeout_method (int): Время ожидания элемента в секундах (по умолчанию: 120).

        Returns:
            bool: True, если прокрутка выполнена успешно.

        """
        recycler = self

        # # Проверка, является ли элемент прокручиваемым
        # if recycler.get_attribute('scrollable') != 'true':
        #     self.logger.error("Элемент не крутиться")
        #     return False

        # Если локатор для прокрутки не указан, используется локатор первого дочернего элемента
        if not locator:
            locator = {'class': self._get_first_child_class()}
        last_child = None
        start_time = time.time()

        # Прокрутка вверх до упора
        while time.time() - start_time < timeout_method:
            try:
                child = self._get_element(locator=locator,
                                          timeout_elem=timeout_elem,
                                          timeout_method=timeout_method,
                                          elements_range=elements_range,
                                          contains=contains,
                                          poll_frequency=poll_frequency,
                                          ignored_exceptions=ignored_exceptions,
                                          )
                if child == last_child:
                    return True
                last_child = child
                self._scroll_up(locator=locator,
                                duration=duration,
                                timeout_elements=timeout_elements,
                                timeout_method=timeout_method,
                                elements_range=elements_range,
                                contains=contains,
                                poll_frequency=poll_frequency,
                                ignored_exceptions=ignored_exceptions,
                                )
            except StaleElementReferenceException:
                continue

        self.logger.error("_scroll_to_top(): Неизвестная ошибка")
        return False

    def _scroll_until_find(self,
                           locator: Union[Tuple, WebElement, 'WebElementExtended', Dict[str, str], str] = None,
                           timeout_elem: float = 10.0,
                           timeout_method: float = 600.0,
                           elements_range: Union[Tuple, typing.List[WebElement], Dict[str, str], None] = None,
                           contains: bool = True,
                           poll_frequency: float = 0.5,
                           ignored_exceptions: typing.Optional[WaitExcTypes] = None,
                           threshold: float = 0.9,
                           check_image: bool = False) -> bool:
        """
        Крутит элемент вниз, а затем вверх для поиска элемента по заданному локатору.

        Args:
            locator (Union[Tuple, WebElement, Dict[str, str], str]): Локатор или элемент, для которого производится
                поиск.
            timeout_method (int): Время на поиск в одном направлении (по умолчанию: 120 вниз и 120 вверх).

        Returns:
            bool: True, если элемент найден. False, если элемент не найден.

        """
        recycler = self

        start_time = time.time()

        last_element_image = None
        last_source = None

        # Прокрутка вниз до поиска элемента
        while time.time() - start_time < timeout_method:
            try:
                if isinstance(locator, str):
                    if self.helpers._is_image_on_the_screen(image=locator, threshold=threshold):
                        return True
                element = self._get_element(locator=locator,
                                            timeout_elem=timeout_elem,
                                            timeout_method=timeout_method,
                                            elements_range=elements_range,
                                            contains=contains,
                                            poll_frequency=poll_frequency,
                                            ignored_exceptions=ignored_exceptions, )
                if element is None:
                    recycler._scroll_down()
                    time.sleep(3)  # чтобы все эффекты прокрутки исчезли
                    if check_image:
                        current_element_image = self.screenshot_as_base64
                        if current_element_image == last_element_image:
                            break
                        last_element_image = self.screenshot_as_base64
                    else:
                        current_source = self.driver.page_source
                        if current_source == last_source:
                            break
                        last_source = self.driver.page_source
                if isinstance(element, WebElement):
                    return True
            except NoSuchElementException:
                continue
            except StaleElementReferenceException:
                element = None
                recycler._scroll_down()
                continue

        # Прокрутка вверх до поиска элемента
        while time.time() - start_time < timeout_method:
            try:
                if isinstance(locator, str):
                    if self.helpers._is_image_on_the_screen(image=locator, threshold=threshold):
                        return True
                element = self._get_element(locator=locator,
                                            timeout_elem=timeout_elem,
                                            timeout_method=timeout_method,
                                            elements_range=elements_range,
                                            contains=contains,
                                            poll_frequency=poll_frequency,
                                            ignored_exceptions=ignored_exceptions, )
                if element is None:
                    recycler._scroll_up()
                    time.sleep(3)  # чтобы все эффекты прокрутки исчезли
                    if check_image:
                        current_element_image = self.screenshot_as_base64
                        if current_element_image == last_element_image:
                            break
                        last_element_image = self.screenshot_as_base64
                    else:
                        current_source = self.driver.page_source
                        if current_source == last_source:
                            break
                        last_source = self.driver.page_source
                if isinstance(element, WebElement):
                    return True
            except NoSuchElementException:
                continue
            except StaleElementReferenceException:
                element = None
                recycler._scroll_up()
                continue

        self.logger.error("_scroll_until_find(): Элемент не найден")
        return False

    def _scroll_and_get(self,
                        locator: Union[Tuple, WebElement, 'WebElementExtended', Dict[str, str], str] = None,
                        timeout_elem: float = 10.0,
                        timeout_method: float = 600.0,
                        elements_range: Union[Tuple, typing.List[WebElement], Dict[str, str], None] = None,
                        contains: bool = True,
                        poll_frequency: float = 0.5,
                        ignored_exceptions: typing.Optional[WaitExcTypes] = None,
                        tries: int = 3,
                        check_image: bool = False
                        ) -> Optional[WebElement]:
        """
        Крутит элемент вниз, а затем вверх для поиска элемента по заданному локатору.

        Args:
            locator (Union[Tuple, WebElement, Dict[str, str], str]): Локатор или элемент, для которого производится
                поиск.
            timeout_method (int): Время на поиск в одном направлении (по умолчанию: 120 вниз и 120 вверх).

        Returns:
            bool: True, если элемент найден. False, если элемент не найден.

        """
        recycler = self

        start_time = time.time()

        last_element_image = None
        last_source = None

        # Прокрутка вниз до поиска элемента
        while time.time() - start_time < timeout_method:
            try:
                if isinstance(locator, str):
                    if self.helpers._is_image_on_the_screen(image=locator):
                        return self._get_element(locator=locator,
                                                 timeout_elem=timeout_elem,
                                                 timeout_method=timeout_method,
                                                 elements_range=elements_range,
                                                 contains=contains,
                                                 poll_frequency=poll_frequency,
                                                 ignored_exceptions=ignored_exceptions, )
                element = self._get_element(locator=locator,
                                            timeout_elem=timeout_elem,
                                            timeout_method=timeout_method,
                                            elements_range=elements_range,
                                            contains=contains,
                                            poll_frequency=poll_frequency,
                                            ignored_exceptions=ignored_exceptions, )
                if element is None or not self._is_within_screen(element):
                    recycler._scroll_down()
                    time.sleep(3)  # чтобы все эффекты прокрутки исчезли
                    if check_image:
                        current_element_image = self.screenshot_as_base64
                        if current_element_image == last_element_image:
                            break
                        last_element_image = self.screenshot_as_base64
                    else:
                        current_source = self.driver.page_source
                        if current_source == last_source:
                            break
                    continue
                if isinstance(element, WebElement):
                    return element
            except NoSuchElementException:
                continue
            except StaleElementReferenceException:
                element = None
                recycler._scroll_down()
                continue

        # Прокрутка вверх до поиска элемента
        while time.time() - start_time < timeout_method:
            try:
                if isinstance(locator, str):
                    if self.helpers._is_image_on_the_screen(image=locator):
                        return self._get_element(locator=locator,
                                                 timeout_elem=timeout_elem,
                                                 timeout_method=timeout_method,
                                                 elements_range=elements_range,
                                                 contains=contains,
                                                 poll_frequency=poll_frequency,
                                                 ignored_exceptions=ignored_exceptions, )
                element = self._get_element(locator=locator,
                                            timeout_elem=timeout_elem,
                                            timeout_method=timeout_method,
                                            elements_range=elements_range,
                                            contains=contains,
                                            poll_frequency=poll_frequency,
                                            ignored_exceptions=ignored_exceptions, )
                if element is None or not self._is_within_screen(element):
                    recycler._scroll_down()
                    time.sleep(3)  # чтобы все эффекты прокрутки исчезли
                    if check_image:
                        current_element_image = self.screenshot_as_base64
                        if current_element_image == last_element_image:
                            break
                        last_element_image = self.screenshot_as_base64
                    else:
                        current_source = self.driver.page_source
                        if current_source == last_source:
                            break
                if isinstance(element, WebElement):
                    return element
            except NoSuchElementException:
                continue
            except StaleElementReferenceException:
                element = None
                recycler._scroll_down()
                continue

        self.logger.error("_scroll_and_get(): Элемент не найден")
        return None



# coding: utf-8
import logging
import time
import typing
from typing import Union, Dict, List, Tuple

import numpy as np
from PIL import Image
from appium.webdriver import WebElement
from selenium.types import WaitExcTypes

from appium_extended.appium_get import AppiumGet


class AppiumWait(AppiumGet):
    """
    Класс расширяющий Appium.
    Обеспечивает ....
    """

    def __init__(self, logger: logging.Logger):
        super().__init__(logger=logger)

    def _wait_for(self,
                  locator: Union[Tuple[str, str], WebElement, 'WebElementExtended', Dict[str, str], str,
                  List[Tuple[str, str]], List[WebElement], List['WebElementExtended'], List[Dict[str, str]], List[
                      str]] = None,
                  image: Union[bytes, np.ndarray, Image.Image, str,
                  List[bytes], List[np.ndarray], List[Image.Image], List[str]] = None,
                  timeout_elem: float = 10.0,
                  timeout_method: float = 600.0,
                  elements_range: Union[Tuple, List[WebElement], Dict[str, str], None] = None,
                  contains: bool = True,
                  sleep: float = 1.0,
                  poll_frequency: float = 0.5,
                  ignored_exceptions: typing.Optional[WaitExcTypes] = None,
                  threshold: float = 0.9,
                  ):
        """
        Ожидает появления на экране указанного локатора или изображения.

        Args:
            locator (Union[Tuple, WebElement, 'WebElementExtended', Dict, str, List], optional):
                - Tuple: локатор в виде кортежа из двух строковых элементов, где первый это стратегия поиска, а второй это селектор, например ("id", "android.widget.ProgressBar").
                - Dict: локатор в виде словаря {'text': 'foo', 'displayed': 'true', 'enabled': 'true'}.
                - str: путь до изображения.
                - List: список из локаторов. Будет ожидание всех элементов из списка.
                По умолчанию None.

            image (Union[bytes, np.ndarray, Image.Image, str, List], optional):
                - bytes: изображение в формате байтов.
                - np.ndarray: изображение в формате массива NumPy.
                - Image.Image: изображение в формате Image (PIL/Pillow).
                - str: путь до файла с изображением.
                По умолчанию None.

            timeout (int, optional): Максимальное время ожидания в секундах. По умолчанию 10.

            contains (bool, optional): Если True, проверяет, содержит ли элемент указанный локатор.
                                       По умолчанию True.

            sleep (int): время ожидания перед новым запросом в процессе поллинга.

        Usages:
            - _wait_for(locator=("id", "android.widget.ProgressBar"), timeout=5)
            - _wait_for(image="path/to/image.png", timeout=10)
            - _wait_for(locator=[("id", "element1"), ("name", "element2")], timeout=5)
            - _wait_for(image=["path/to/image1.png", "path/to/image2.png"], timeout=10)


        Returns:
            bool: True, если элементы или изображения найдены в течение заданного времени, иначе False.

        Raises:
            None: Метод не вызывает исключений.

        Notes:
            - Метод использует внутренние функции для поиска элементов и изображений.
            - Параметр `contains` используется только при поиске по локатору.
        """
        if locator is not None:
            if not isinstance(locator, List):
                locator = [locator]

            # Loop through each locator
            for i in locator:
                # Check if the element is present
                if self._get_element(locator=i,
                                     timeout_elem=timeout_elem,
                                     timeout_method=timeout_method,
                                     elements_range=elements_range,
                                     contains=contains,
                                     poll_frequency=poll_frequency,
                                     ignored_exceptions=ignored_exceptions,) is None:
                    return False

        if image is not None:
            start_time = time.time()
            if not isinstance(image, List):
                image = [image]

            # Loop through each image
            for i in image:
                # Check if the image is on the screen within the timeout period
                while not self.helpers._is_image_on_the_screen(image=i, threshold=threshold) and time.time() - start_time < timeout_method:
                    time.sleep(sleep)
                if not self.helpers._is_image_on_the_screen(image=i, threshold=threshold):
                    return False

        # Return True if all conditions are met
        return True

    def _wait_for_not(self,
                      locator: Union[Tuple[str, str], WebElement, 'WebElementExtended', Dict[str, str], str,
                      List[Tuple[str, str]], List[WebElement], List['WebElementExtended'], List[Dict[str, str]], List[
                          str]] = None,
                      image: Union[bytes, np.ndarray, Image.Image, str,
                      List[bytes], List[np.ndarray], List[Image.Image], List[str]] = None,
                      timeout_elem: float = 10.0,
                      timeout_method: float = 600.0,
                      elements_range: Union[Tuple, List[WebElement], Dict[str, str], None] = None,
                      contains: bool = True,
                      poll_frequency: float = 0.5,
                      ignored_exceptions: typing.Optional[WaitExcTypes] = None,
                      sleep: float = 1.0,
                      threshold: float = 0.9,
                      ):
        """
        Ожидает пока указанный локатор или изображение исчезнет с экрана или DOM.

        Args:
            locator (Union[Tuple[str, str], WebElement, 'WebElementExtended', Dict[str, str], str,
                           List[Tuple[str, str]], List[WebElement], List['WebElementExtended'],
                           List[Dict[str, str]], List[str]], optional): The locator(s) to wait for.
                           Can be a single locator or a list of locators. Defaults to None.

            image (Union[bytes, np.ndarray, Image.Image, str,
                         List[bytes], List[np.ndarray], List[Image.Image], List[str]], optional):
                         The image(s) to wait for. Can be a single image or a list of images.
                         Defaults to None.

            timeout (float, optional): The maximum time to wait in seconds. Defaults to 10.0.

            contains (bool, optional): If True, checks if the element contains the specified locator.
                                       If False, checks if the element exactly matches the specified locator.
                                       Defaults to True.

            sleep (int): время ожидания перед новым запросом в процессе поллинга.

        Returns:
            bool: True if the element(s) are found within the timeout period, False otherwise.
        """
        if locator is not None:
            if not isinstance(locator, List):
                locator = [locator]

            # Loop through each locator
            start_time = time.time()
            while time.time() - start_time < timeout_method:
                locators_present = False
                for i in locator:
                    # Check if the element is present
                    if not self._get_element(locator=i,
                                             timeout_elem=timeout_elem,
                                             timeout_method=timeout_method,
                                             elements_range=elements_range,
                                             contains=contains,
                                             poll_frequency=poll_frequency,
                                             ignored_exceptions=ignored_exceptions,) is None:
                        locators_present = True
                if not locators_present:
                    return True
                time.sleep(sleep)
            raise TimeoutError

        if image is not None:
            if not isinstance(image, List):
                image = [image]

            # Loop through each image
            start_time = time.time()
            while time.time() - start_time < timeout_method:
                images_present = False
                for i in image:
                    # Check if the image is on the screen within the timeout period
                    if self.helpers._is_image_on_the_screen(image=i, threshold=threshold):
                        images_present = True
                if not images_present:
                    return True
                time.sleep(sleep)
            raise TimeoutError
        return False

    @staticmethod
    def _wait_return_true(method, timeout: float = 10.0, sleep: float = 1.0, *args, **kwargs):
        """
        Ожидает пока метод не вернет True.
        Args:
            method: ссылка на метод
            timeout (float): таймаут на ожидание
            sleep (int): время ожидания перед новым запросом в процессе поллинга.
        """
        start_time = time.time()
        while time.time() - start_time < timeout:
            if method(*args, **kwargs):
                return
            time.sleep(sleep)
        raise TimeoutError





from appium_extended_web_element.web_element_get import WebElementGet


class WebElementIs(WebElementGet):
    """
    Класс для выполнения действий нажатия (Tap), а также нажатия и перемещения с использованием элементов веб-страницы.
    Наследуется от класса WebElementGet.
    """

    def __init__(self, base, element_id):
        super().__init__(base=base, element_id=element_id)

    def _is_within_screen(self, element):
        screen_size = self.terminal.get_screen_resolution()  # Получаем размеры экрана
        screen_width = screen_size[0]  # Ширина экрана
        screen_height = screen_size[1]  # Высота экрана
        element_location = element.location  # Получаем координаты элемента
        element_size = element.size  # Получаем размеры элемента
        if (
                element_location['y'] + element_size['height'] > screen_height or
                element_location['x'] + element_size['width'] > screen_width or
                element_location['y'] < 0 or
                element_location['x'] < 0
        ):
            # Если элемент находится за пределами экрана
            return False
        # Если элемент находится на экране
        return True



import subprocess

from setuptools import setup, find_packages

# Общее описание пакета
with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


VERSION = '0.16.157'


setup(
    name='AppiumExtended',
    version=VERSION,
    description='An extension library for adding ease of use Appium-Python-Client',
    long_description=long_description,
    author='molokov-klim',
    packages=find_packages(),
    install_requires=[
        'Appium-Python-Client==2.11.1',
        'allure-pytest==2.13.2',
        "zlib-compress==0.0.1",
        "zlib-decompress==0.0.2",
        "pylibjpeg==1.4.0",
        'Pillow==9.5.0',
        'requests==2.31.0',
        'pyserial==3.5',
        'opencv-python==4.8.0.74',
        'pytesseract==0.3.10',
        'numpy==1.25.1',
        'paramiko==3.4.0',
        'scp==0.14.5'
    ],
    long_description_content_type='text/markdown',  # Указываем тип контента (обычный текст)
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    url='https://github.com/molokov-klim/appium_extended',
)
